#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Any, Dict

from classy_vision import tasks


class ClassyHookFunctions(Enum):
    """
    Enumeration of all the hook functions in the ClassyHook class.
    """

    on_rendezvous = auto()
    on_start = auto()
    on_phase_start = auto()
    on_sample = auto()
    on_forward = auto()
    on_loss_and_meter = auto()
    on_backward = auto()
    on_update = auto()
    on_phase_end = auto()
    on_end = auto()


class ClassyHookState:
    """Class to store state within instances of ClassyHook.

    Any serializable data can be stored in the instance's attributes.
    """

    def get_classy_state(self) -> Dict[str, Any]:
        return self.__dict__

    def set_classy_state(self, state_dict: Dict[str, Any]):
        self.__dict__ = state_dict


class ClassyHook(ABC):
    """Base class for hooks.

    Hooks allow to inject behavior at different places of the training loop, which
        are listed below in the chronological order.

    on_start -> on_phase_start -> on_sample -> on_forward -> on_loss_and_meter ->
        on_backward -> on_update -> on_phase_end -> on_end

    Deriving classes should call super().__init__() and store any state in
    self.state. Any state added to this property should be serializable.
    E.g. -
    class MyHook(ClassyHook):
        def __init__(self, a, b):
            super().__init__()
            self.state.a = [1,2,3]
            self.state.b = "my_hook"
            # the following line is not allowed
            # self.state.my_lambda = lambda x: x^2

    """

    def __init__(self):
        self.state = ClassyHookState()

    def _noop(self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]) -> None:
        """Derived classes can set their hook functions to this.

        This is useful if they want those hook functions to not do anything.

        """
        pass

    @classmethod
    def name(cls) -> str:
        """Returns the name of the class."""
        return cls.__name__

    @abstractmethod
    def on_rendezvous(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Called when the trainers rendezvous."""
        pass

    @abstractmethod
    def on_start(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Called at the start of training."""
        pass

    @abstractmethod
    def on_phase_start(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Called at the start of each phase."""
        pass

    @abstractmethod
    def on_sample(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Called each time trainer obtained a sample from the dataset."""
        pass

    @abstractmethod
    def on_forward(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Called each time forward pass is done in the model."""
        pass

    @abstractmethod
    def on_loss_and_meter(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Called each time after a loss has been computed and meters are updated."""
        pass

    @abstractmethod
    def on_backward(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Called each time a backward step is performed on the loss."""
        pass

    @abstractmethod
    def on_update(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Called each time after parameters have been updated by the optimizer."""
        pass

    @abstractmethod
    def on_phase_end(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Called at the end of each phase (epoch)."""
        pass

    @abstractmethod
    def on_end(self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]) -> None:
        """Called at the end of training."""
        pass

    def get_classy_state(self) -> Dict[str, Any]:
        """Get the state of the ClassyHook.

        The returned state is used for checkpointing.

        Returns:
            A state dictionary containing the state of the hook.\

        """
        return self.state.get_classy_state()

    def set_classy_state(self, state_dict: Dict[str, Any]) -> None:
        """Set the state of the ClassyHook.

        Args:
            state_dict: The state dictionary. Must be the output of a call to
                :method:`get_classy_state`.

        This is used to load the state of the hook from a checkpoint.

        """
        self.state.set_classy_state(state_dict)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from typing import Any, Dict

import numpy as np
import torch
import torch.nn.functional as F
from classy_vision.losses import ClassyLoss, register_loss


@register_loss("soft_target_cross_entropy")
class SoftTargetCrossEntropyLoss(ClassyLoss):
    def __init__(self, ignore_index, reduction, normalize_targets):
        """Intializer for the soft target cross-entropy loss loss.
        This allows the targets for the cross entropy loss to be multilabel

        Config params:
        'weight': weight of sample (not yet implemented),
        'ignore_index': sample should be ignored for loss (optional),
        'reduction': specifies reduction to apply to the output (optional),
        """
        super(SoftTargetCrossEntropyLoss, self).__init__()
        self._ignore_index = ignore_index
        self._reduction = reduction
        assert normalize_targets in [None, "count_based"]
        self._normalize_targets = normalize_targets
        if self._reduction != "mean":
            raise NotImplementedError(
                'reduction type "{}" not implemented'.format(self._reduction)
            )
        self._eps = np.finfo(np.float32).eps

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "SoftTargetCrossEntropyLoss":
        """Instantiates a SoftTargetCrossEntropyLoss from a configuration.

        Args:
            config: A configuration for a SoftTargetCrossEntropyLoss.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A SoftTargetCrossEntropyLoss instance.
        """

        if "weight" in config:
            raise NotImplementedError('"weight" not implemented')
        return cls(
            ignore_index=config.get("ignore_index", -100),
            reduction=config.get("reduction", "mean"),
            normalize_targets=config.get("normalize_targets", "count_based"),
        )

    def forward(self, output, target):
        """for N examples and C classes
        - output: N x C these are raw outputs (without softmax/sigmoid)
        - target: N x C corresponding targets

        Target elements set to ignore_index contribute 0 loss.

        Samples where all entries are ignore_index do not contribute to the loss
        reduction.
        """
        assert (
            output.shape == target.shape
        ), "SoftTargetCrossEntropyLoss requires output and target to be same"
        valid_mask = target != self._ignore_index
        valid_targets = target.float() * valid_mask.float()
        if self._normalize_targets == "count_based":
            valid_targets /= self._eps + valid_targets.sum(dim=1, keepdim=True)
        per_sample_per_target_loss = -valid_targets * F.log_softmax(output, -1)
        # perform reduction
        if self._reduction == "mean":
            per_sample_loss = torch.sum(per_sample_per_target_loss, -1)
            # normalize based on the number of samples with > 0 non-ignored targets
            loss = per_sample_loss.sum() / torch.sum(
                (torch.sum(valid_mask, -1) > 0)
            ).clamp(min=1)
        return loss



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, Optional

import torch.nn as nn


class ClassyHead(nn.Module):
    """
    Base class for heads that can be attached to :class:`ClassyModel`.

    A head is a regular :class:`torch.nn.Module` that can be attached to a
    pretrained model. This enables a form of transfer learning: utilizing a
    model trained for one dataset to extract features that can be used for
    other problems. A head must be attached to a :class:`ClassyBlock` within a
    :class:`ClassyModel`.
    """

    def __init__(
        self, unique_id: Optional[str] = None, num_classes: Optional[int] = None
    ):
        """
        Constructs a ClassyHead.

        Args:
            unique_id: A unique identifier for the head. Multiple instances of
                the same head might be attached to a model, and unique_id is used
                to refer to them.

            num_classes: Number of classes for the head.
        """
        super().__init__()
        self.unique_id = unique_id or self.__class__.__name__
        self.num_classes = num_classes

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ClassyHead":
        """Instantiates a ClassyHead from a configuration.

        Args:
            config: A configuration for the ClassyHead.

        Returns:
            A ClassyHead instance.
        """
        raise NotImplementedError

    def forward(self, x):
        """
        Performs inference on the head.

        This is a regular PyTorch method, refer to :class:`torch.nn.Module` for
        more details
        """
        raise NotImplementedError



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest
import unittest.mock as mock
from test.generic.utils import compare_batches, compare_samples

import classy_vision.dataset.classy_dataset as classy_dataset
import torch
from classy_vision.dataset import build_dataset, register_dataset
from classy_vision.dataset.core import ListDataset
from torch.utils.data import DataLoader


DUMMY_SAMPLES_1 = [
    {"input": torch.tensor([[[0, 1], [2, 3]]]), "target": torch.tensor([[0]])}
]


DUMMY_SAMPLES_2 = [
    {"input": torch.tensor([[[0, 1], [2, 3]]]), "target": torch.tensor([[0]])},
    {"input": torch.tensor([[[4, 5], [6, 7]]]), "target": torch.tensor([[1]])},
]

BATCHED_DUMMY_SAMPLES_2 = [
    {
        "input": torch.tensor([[[[0, 1], [2, 3]]], [[[4, 5], [6, 7]]]]),
        "target": torch.tensor([[[0]], [[1]]]),
    }
]

DUMMY_CONFIG = {"name": "test_dataset", "dummy0": 0, "dummy1": 1}

OTHER_DUMMY_CONFIG = {"name": "other_test_dataset", "dummy0": 0, "dummy1": 1}


def mock_get_world_size():
    return 2


def mock_get_rank():
    return 1


@register_dataset("test_dataset")
class TestDataset(classy_dataset.ClassyDataset):
    """Test dataset for validating registry functions"""

    def __init__(
        self,
        samples,
        batchsize_per_replica=1,
        num_samples=None,
        shuffle=False,
        transform=None,
    ):
        input_tensors = [sample["input"] for sample in samples]
        target_tensors = [sample["target"] for sample in samples]
        dataset = ListDataset(input_tensors, target_tensors, loader=lambda x: x)
        super().__init__(
            dataset=dataset,
            split=None,
            batchsize_per_replica=batchsize_per_replica,
            shuffle=shuffle,
            transform=transform,
            num_samples=len(samples) if num_samples is None else num_samples,
        )

    @classmethod
    def from_config(cls, config, *args, **kwargs):
        return cls(*args, **kwargs)


@register_dataset("other_test_dataset")
class OtherTestDataset(classy_dataset.ClassyDataset):
    """
    Test dataset for validating registry functions that has a different
    type than TestDataset
    """

    def __init__(self, samples, batchsize_per_replica=1):
        input_tensors = [sample["input"] for sample in samples]
        target_tensors = [sample["target"] for sample in samples]
        dataset = ListDataset(input_tensors, target_tensors, loader=lambda x: x)
        super().__init__(
            dataset=dataset,
            split=None,
            batchsize_per_replica=batchsize_per_replica,
            shuffle=False,
            transform=None,
            num_samples=len(samples),
        )

    @classmethod
    def from_config(cls, config, *args, **kwargs):
        return cls(*args, **kwargs)


class TestRegistryFunctions(unittest.TestCase):
    """
    Tests functions that use registry
    """

    def test_build_model(self):
        dataset = build_dataset(DUMMY_CONFIG, DUMMY_SAMPLES_1)
        self.assertTrue(isinstance(dataset, TestDataset))


class TestClassyDataset(unittest.TestCase):
    """
    Tests member functions of ClassyDataset. Note, NotImplemented
    functions are mocked in TestDataset class.
    """

    def setUp(self):
        self.dataset1 = build_dataset(DUMMY_CONFIG, DUMMY_SAMPLES_1)
        self.dataset2 = build_dataset(DUMMY_CONFIG, DUMMY_SAMPLES_2)

    def _compare_samples(self, sample1, sample2):
        compare_samples(self, sample1, sample2)

    def _compare_batches(self, batch1, batch2):
        compare_batches(self, batch1, batch2)

    def test_init(self):
        self.assertTrue(self.dataset1 is not None)
        self.assertTrue(self.dataset2 is not None)

    def test_len(self):
        self.assertEqual(len(self.dataset1), 1)
        self.assertEqual(len(self.dataset2), 2)

    def test_getitem(self):
        sample = self.dataset1[0]
        self._compare_samples(sample, DUMMY_SAMPLES_1[0])

        for idx in range(len(self.dataset2)):
            sample = self.dataset2[idx]
            self._compare_samples(sample, DUMMY_SAMPLES_2[idx])

    def test_get_iterator(self):
        # Verifies that we can retrieve samples with iterators
        dl = self.dataset1.iterator(num_workers=0)
        assert isinstance(
            dl, DataLoader
        ), "Classy Iterator should return instance of PyTorch Dataloader"
        next(iter(dl))

        dl = self.dataset1.iterator(num_workers=2)
        assert isinstance(
            dl, DataLoader
        ), "Classy Iterator should return instance of PyTorch Dataloader"
        it = iter(dl)
        next(it)
        # Because we use multiprocessing we delete the iterable to
        # shutdown workers
        del it

    def test_batch_logic(self):
        dataset = TestDataset(DUMMY_SAMPLES_2, batchsize_per_replica=2)
        dl = dataset.iterator(num_workers=0)
        batch = next(iter(dl))
        self.assertEqual(batch["input"].size()[0], 2)
        self._compare_batches(batch, BATCHED_DUMMY_SAMPLES_2[0])

    @mock.patch(
        "classy_vision.dataset.classy_dataset.get_world_size", mock_get_world_size
    )
    @mock.patch("classy_vision.dataset.classy_dataset.get_rank", mock_get_rank)
    def test_shard_logic(self):
        # This test uses a world size of 2, rank 1 to verify that the
        # second sample is returned by the dataloader
        dataset = TestDataset(DUMMY_SAMPLES_2, batchsize_per_replica=1)
        dl = dataset.iterator(num_workers=0)
        sample = next(iter(dl))
        self._compare_batches(sample, DUMMY_SAMPLES_2[1])

    def test_num_samples_logic(self):
        dataset = TestDataset(DUMMY_SAMPLES_2)
        self.assertEqual(len(dataset), 2)

        dataset = TestDataset(DUMMY_SAMPLES_2, num_samples=1)
        # Verify len returns right value for dataset
        self.assertEqual(len(dataset), 1)
        # Verify len returns right value for iterator
        self.assertEqual(len(dataset.iterator(num_workers=0)), 1)
        # Verify iterator returns correct number of samples
        it = iter(dataset.iterator(num_workers=0))
        num_samples = 0
        while True:
            try:
                next(it)
                num_samples += 1
            except StopIteration:
                break
        self.assertEqual(num_samples, 1)

        # Check assert for num_samples > length of base dataset
        dataset = TestDataset(DUMMY_SAMPLES_2, num_samples=3)
        with self.assertRaises(AssertionError):
            len(dataset)

    def test_shuffle_logic(self):
        # Simple samples to test shuffling, just a single value tensor
        # so we know how things were shuffled
        dummy_samples_10 = [
            {"input": torch.tensor([[0]]), "target": torch.tensor([0])},
            {"input": torch.tensor([[1]]), "target": torch.tensor([0])},
            {"input": torch.tensor([[2]]), "target": torch.tensor([0])},
            {"input": torch.tensor([[3]]), "target": torch.tensor([0])},
            {"input": torch.tensor([[4]]), "target": torch.tensor([0])},
            {"input": torch.tensor([[5]]), "target": torch.tensor([0])},
            {"input": torch.tensor([[6]]), "target": torch.tensor([0])},
            {"input": torch.tensor([[7]]), "target": torch.tensor([0])},
            {"input": torch.tensor([[8]]), "target": torch.tensor([0])},
            {"input": torch.tensor([[9]]), "target": torch.tensor([0])},
        ]
        dataset = TestDataset(dummy_samples_10, shuffle=True)

        def unpack_tensors(tensor_list):
            return [t["input"].item() for t in tensor_list]

        # Epoch 0
        iterator = dataset.iterator(num_workers=0, current_phase_id=0)
        it = iter(iterator)
        epoch_0_list = [sample for sample in it]
        epoch_0_list = unpack_tensors(epoch_0_list)

        # Epoch 1
        iterator = dataset.iterator(num_workers=0, current_phase_id=1)
        it = iter(iterator)
        epoch_1_list = [sample for sample in it]
        epoch_1_list = unpack_tensors(epoch_1_list)

        # Should be same length, should be shuffled, should be
        # different shuffles for each epoch
        self.assertEqual(len(epoch_0_list), len(epoch_1_list))
        self.assertTrue(epoch_0_list != [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        self.assertTrue(epoch_0_list != epoch_1_list)

        # Test different shuffle seeds
        iterator = dataset.iterator(num_workers=0, current_phase_id=0, shuffle_seed=10)
        it = iter(iterator)
        epoch_0_seed_10_list = [sample for sample in it]
        epoch_0_seed_10_list = unpack_tensors(epoch_0_seed_10_list)
        self.assertTrue(epoch_0_seed_10_list != epoch_0_list)

    def test_transform_logic(self):
        def _return_1_transform(sample):
            return 1

        dataset = TestDataset(DUMMY_SAMPLES_2, transform=_return_1_transform)
        sample = dataset[0]
        self.assertEqual(sample, 1)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Any, Callable, Dict, List, Optional

import torch
from torchvision.datasets.kinetics import Kinetics400

from . import register_dataset
from .classy_video_dataset import ClassyVideoDataset
from .transforms.util_video import build_video_field_transform_default


@register_dataset("kinetics400")
class Kinetics400Dataset(ClassyVideoDataset):
    """Kinetics-400 is an action recognition video dataset, and it has 400 classes.

    Page: <https://deepmind.com/research/open-source/open-source-datasets/kinetics/>
    It is originally published in (https://arxiv.org/pdf/1705.06950.pdf).

    This dataset consider every video as a collection of video clips of fixed size,
    specified by ``frames_per_clip``, where the step in frames between each clip
    is given by ``step_between_clips``. It uses clip sampler to sample clips
    from each video. For training set, a random clip sampler is used to
    sample a small number of clips (e.g. 1) from each video
    For testing set, a uniform clip sampler is used to evenly sample a large
    number of clips (e.g. 10) from the video.

    To give an example, for 2 videos with 10 and 15 frames respectively, if
    ``frames_per_clip=5`` and ``step_between_clips=5``, the dataset size
    will be (2 + 3) = 5, where the first two elements will come from video 1,
    and the next three elements from video 2. Note that we drop clips which do
    not have exactly ``frames_per_clip`` elements, so not all frames in a video
    might be present.

    We assume videos are already trimmed to 10-second clip, and are stored in a
    folder.

    It is built on top of Kinetics400 dataset class in TorchVision.

    """

    def __init__(
        self,
        split: str,
        batchsize_per_replica: int,
        shuffle: bool,
        transform: Callable,
        num_samples: Optional[int],
        frames_per_clip: int,
        video_width: int,
        video_height: int,
        video_min_dimension: int,
        audio_samples: int,
        audio_channels: int,
        step_between_clips: int,
        frame_rate: Optional[int],
        clips_per_video: int,
        video_dir: str,
        extensions: List[str],
        metadata_filepath: str,
    ):
        """The constructor of Kinetics400Dataset.

        Args:
            split: dataset split which can be either "train" or "test"
            batchsize_per_replica: batch size per model replica
            shuffle: If true, shuffle the dataset
            transform: a dict where transforms video and audio data
            num_samples: if provided, it will subsample dataset
            frames_per_clip: the No. of frames in a video clip
            video_width: rescaled video width. If 0, keep original width
            video_height: rescaled video height. If 0, keep original height
            video_min_dimension: rescale video so that min(height, width) =
                video_min_dimension. If 0, keep original video resolution. Note
                only one of (video_width, video_height) and (video_min_dimension)
                can be set
            audio_samples: desired audio sample rate. If 0, keep original
                audio sample rate
            audio_channels: desire No. of audio channel. If 0, keep original audio
                channels
            step_between_clips: No. of frames between each clip.
            frame_rate: desired video frame rate. If None, keep
                orignal video frame rate.
            clips_per_video: No. of clips to sample from each video
            video_dir: path to video folder
            extensions: A list of file extensions, such as "avi" and "mp4". Only
                video matching those file extensions are added to the dataset
            metadata_filepath: path to the dataset meta data

        """
        # dataset metadata includes the path of video file, the pts of frames in
        # the video and other meta info such as video fps, duration, audio sample rate.
        # Users do not need to know the details of metadata. The computing, loading
        # and saving logic of metata are all handled inside of the dataset.
        # Given the "metadata_file" path, if such file exists, we load it as meta data.
        # Otherwise, we compute the meta data, and save it at "metadata_file" path.
        metadata = None
        if os.path.exists(metadata_filepath):
            metadata = Kinetics400Dataset.load_metadata(
                metadata_filepath, video_dir=video_dir, update_file_path=True
            )

        dataset = Kinetics400(
            video_dir,
            frames_per_clip,
            step_between_clips=step_between_clips,
            frame_rate=frame_rate,
            _precomputed_metadata=metadata,
            extensions=extensions,
            num_workers=torch.get_num_threads(),
            _video_width=video_width,
            _video_height=video_height,
            _video_min_dimension=video_min_dimension,
            _audio_samples=audio_samples,
            _audio_channels=audio_channels,
        )
        metadata = dataset.metadata
        if metadata and not os.path.exists(metadata_filepath):
            Kinetics400Dataset.save_metadata(metadata, metadata_filepath)

        super().__init__(
            dataset,
            split,
            batchsize_per_replica,
            shuffle,
            transform,
            num_samples,
            clips_per_video,
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "Kinetics400Dataset":
        """Instantiates a UCF101Dataset from a configuration.

        Args:
            config: A configuration for a Kinetics400Dataset.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A Kinetics400Dataset instance.
        """
        required_args = ["split", "metadata_file", "video_dir"]
        assert all(
            arg in config for arg in required_args
        ), f"The arguments {required_args} are all required."

        split = config["split"]
        audio_channels = config.get("audio_channels", 0)
        (
            transform_config,
            batchsize_per_replica,
            shuffle,
            num_samples,
            frames_per_clip,
            video_width,
            video_height,
            video_min_dimension,
            audio_samples,
            step_between_clips,
            frame_rate,
            clips_per_video,
        ) = cls.parse_config(config)
        extensions = config.get("extensions", ("mp4"))

        transform = build_video_field_transform_default(transform_config, split)

        return cls(
            split,
            batchsize_per_replica,
            shuffle,
            transform,
            num_samples,
            frames_per_clip,
            video_width,
            video_height,
            video_min_dimension,
            audio_samples,
            audio_channels,
            step_between_clips,
            frame_rate,
            clips_per_video,
            config["video_dir"],
            extensions,
            config["metadata_file"],
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from classy_vision.tasks import build_task


def get_test_task_config(head_num_classes=1000):
    return {
        "name": "classification_task",
        "num_epochs": 12,
        "loss": {"name": "CrossEntropyLoss"},
        "dataset": {
            "train": {
                "name": "synthetic_image",
                "split": "train",
                "crop_size": 224,
                "class_ratio": 0.5,
                "num_samples": 2000,
                "seed": 0,
                "batchsize_per_replica": 32,
                "use_shuffle": True,
                "transforms": [
                    {
                        "name": "apply_transform_to_key",
                        "transforms": [
                            {"name": "ToTensor"},
                            {
                                "name": "Normalize",
                                "mean": [0.485, 0.456, 0.406],
                                "std": [0.229, 0.224, 0.225],
                            },
                        ],
                        "key": "input",
                    }
                ],
            },
            "test": {
                "name": "synthetic_image",
                "split": "test",
                "crop_size": 224,
                "class_ratio": 0.5,
                "num_samples": 2000,
                "seed": 0,
                "batchsize_per_replica": 32,
                "use_shuffle": False,
                "transforms": [
                    {
                        "name": "apply_transform_to_key",
                        "transforms": [
                            {"name": "ToTensor"},
                            {
                                "name": "Normalize",
                                "mean": [0.485, 0.456, 0.406],
                                "std": [0.229, 0.224, 0.225],
                            },
                        ],
                        "key": "input",
                    }
                ],
            },
        },
        "meters": {"accuracy": {"topk": [1, 5]}},
        "model": {
            "name": "resnet",
            "num_blocks": [3, 4, 6, 3],
            "small_input": False,
            "zero_init_bn_residuals": True,
            "heads": [
                {
                    "name": "fully_connected",
                    "unique_id": "default_head",
                    "num_classes": head_num_classes,
                    "fork_block": "block3-2",
                    "in_plane": 2048,
                }
            ],
        },
        "optimizer": {
            "name": "sgd",
            "num_epochs": 12,
            "lr": {"name": "step", "values": [0.1, 0.01]},
            "weight_decay": 1e-4,
            "momentum": 0.9,
        },
    }


def get_fast_test_task_config(head_num_classes=1000):
    return {
        "name": "classification_task",
        "num_epochs": 1,
        "loss": {"name": "CrossEntropyLoss"},
        "dataset": {
            "train": {
                "name": "synthetic_image",
                "split": "train",
                "crop_size": 20,
                "class_ratio": 0.5,
                "num_samples": 10,
                "seed": 0,
                "batchsize_per_replica": 2,
                "use_shuffle": False,
                "transforms": [
                    {
                        "name": "apply_transform_to_key",
                        "transforms": [
                            {"name": "ToTensor"},
                            {
                                "name": "Normalize",
                                "mean": [0.485, 0.456, 0.406],
                                "std": [0.229, 0.224, 0.225],
                            },
                        ],
                        "key": "input",
                    }
                ],
            },
            "test": {
                "name": "synthetic_image",
                "split": "test",
                "crop_size": 20,
                "class_ratio": 0.5,
                "num_samples": 10,
                "seed": 0,
                "batchsize_per_replica": 2,
                "use_shuffle": False,
                "transforms": [
                    {
                        "name": "apply_transform_to_key",
                        "transforms": [
                            {"name": "ToTensor"},
                            {
                                "name": "Normalize",
                                "mean": [0.485, 0.456, 0.406],
                                "std": [0.229, 0.224, 0.225],
                            },
                        ],
                        "key": "input",
                    }
                ],
            },
        },
        "model": {
            "name": "resnet",
            "num_blocks": [1],
            "small_input": False,
            "zero_init_bn_residuals": True,
            "heads": [
                {
                    "name": "fully_connected",
                    "unique_id": "default_head",
                    "num_classes": head_num_classes,
                    "fork_block": "block0-0",
                    "in_plane": 256,
                }
            ],
        },
        "meters": {"accuracy": {"topk": [1]}},
        "optimizer": {"name": "sgd", "lr": 0.01, "weight_decay": 1e-4, "momentum": 0.9},
    }


def get_test_classy_task():
    config = get_test_task_config()
    task = build_task(config)
    return task


def get_test_mlp_task_config():
    return {
        "name": "classification_task",
        "num_epochs": 10,
        "loss": {"name": "CrossEntropyLoss"},
        "dataset": {
            "train": {
                "name": "synthetic_image",
                "split": "train",
                "num_classes": 2,
                "crop_size": 20,
                "class_ratio": 0.5,
                "num_samples": 10,
                "seed": 0,
                "batchsize_per_replica": 3,
                "use_augmentation": False,
                "use_shuffle": True,
                "transforms": [
                    {
                        "name": "apply_transform_to_key",
                        "transforms": [
                            {"name": "ToTensor"},
                            {
                                "name": "Normalize",
                                "mean": [0.485, 0.456, 0.406],
                                "std": [0.229, 0.224, 0.225],
                            },
                        ],
                        "key": "input",
                    }
                ],
            },
            "test": {
                "name": "synthetic_image",
                "split": "test",
                "num_classes": 2,
                "crop_size": 20,
                "class_ratio": 0.5,
                "num_samples": 10,
                "seed": 0,
                "batchsize_per_replica": 1,
                "use_augmentation": False,
                "use_shuffle": False,
                "transforms": [
                    {
                        "name": "apply_transform_to_key",
                        "transforms": [
                            {"name": "ToTensor"},
                            {
                                "name": "Normalize",
                                "mean": [0.485, 0.456, 0.406],
                                "std": [0.229, 0.224, 0.225],
                            },
                        ],
                        "key": "input",
                    }
                ],
            },
        },
        "model": {
            "name": "mlp",
            # 3x20x20 = 1200
            "input_dim": 1200,
            "output_dim": 1000,
            "hidden_dims": [10],
        },
        "meters": {"accuracy": {"topk": [1]}},
        "optimizer": {
            "name": "sgd",
            "num_epochs": 10,
            "lr": 0.1,
            "weight_decay": 1e-4,
            "momentum": 0.9,
        },
    }


def get_test_model_configs():
    return [
        # resnet 101
        {
            "name": "resnet",
            "num_blocks": [3, 4, 6, 3],
            "small_input": False,
            "heads": [
                {
                    "name": "fully_connected",
                    "unique_id": "default_head",
                    "num_classes": 1000,
                    "fork_block": "block3-2",
                    "in_plane": 2048,
                }
            ],
        },
        # resnext 101 32-4d
        {
            "name": "resnext",
            "num_blocks": [3, 4, 6, 3],
            "base_width_and_cardinality": [4, 32],
            "small_input": False,
            "heads": [
                {
                    "name": "fully_connected",
                    "unique_id": "default_head",
                    "num_classes": 1000,
                    "fork_block": "block3-2",
                    "in_plane": 2048,
                }
            ],
        },
    ]


def get_test_video_task_config():
    return {
        "name": "classification_task",
        "num_epochs": 27,
        "loss": {"name": "CrossEntropyLoss"},
        "dataset": {
            "train": {
                "name": "synthetic_video",
                "split": "train",
                "batchsize_per_replica": 8,
                "use_shuffle": True,
                "num_samples": 128,
                "frames_per_clip": 8,
                "video_height": 128,
                "video_width": 160,
                "num_classes": 50,
                "clips_per_video": 1,
            },
            "test": {
                "name": "synthetic_video",
                "split": "test",
                "batchsize_per_replica": 10,
                "use_shuffle": False,
                "num_samples": 40,
                "frames_per_clip": 8,
                "video_height": 128,
                "video_width": 160,
                "num_classes": 50,
                "clips_per_video": 10,
            },
        },
        "meters": {"accuracy": {"topk": [1, 5]}},
        "model": {
            "name": "resnext3d",
            "frames_per_clip": 8,
            "input_planes": 3,
            "clip_crop_size": 224,
            "skip_transformation_type": "postactivated_shortcut",
            "residual_transformation_type": "postactivated_bottleneck_transformation",
            "num_blocks": [3, 4, 6, 3],
            "input_key": "video",
            "stem_name": "resnext3d_stem",
            "stem_planes": 64,
            "stem_temporal_kernel": 5,
            "stem_spatial_kernel": 7,
            "stem_maxpool": True,
            "stage_planes": 64,
            "stage_temporal_kernel_basis": [[3], [3, 1], [3, 1], [1, 3]],
            "temporal_conv_1x1": [True, True, True, True],
            "stage_temporal_stride": [1, 1, 1, 1],
            "stage_spatial_stride": [1, 2, 2, 2],
            "num_groups": 1,
            "width_per_group": 64,
            "num_classes": 50,
            "heads": [
                {
                    "name": "fully_convolutional_linear",
                    "unique_id": "default_head",
                    "pool_size": [8, 7, 7],
                    "activation_func": "softmax",
                    "num_classes": 50,
                    "fork_block": "pathway0-stage4-block2",
                    "in_plane": 512,
                    "use_dropout": True,
                }
            ],
        },
        "optimizer": {
            "name": "sgd",
            "lr": {
                "name": "multistep",
                "num_epochs": 10,
                "values": [0.1, 0.01, 0.001, 0.0001],
                "milestones": [3, 7, 9],
            },
            "weight_decay": 0.0001,
            "momentum": 0.9,
        },
    }


def get_test_classy_video_task():
    config = get_test_video_task_config()
    task = build_task(config)
    return task



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path

from classy_vision.generic.registry_utils import import_all_modules

from .classy_dataset import ClassyDataset


FILE_ROOT = Path(__file__).parent

DATASET_REGISTRY = {}
DATASET_CLASS_NAMES = set()


def build_dataset(config, *args, **kwargs):
    """Builds a ClassyDataset from a config.

    This assumes a 'name' key in the config which is used to determine what
    dataset class to instantiate. For instance, a config `{"name": "my_dataset",
    "folder": "/data"}` will find a class that was registered as "my_dataset"
    (see :func:`register_dataset`) and call .from_config on it."""
    return DATASET_REGISTRY[config["name"]].from_config(config, *args, **kwargs)


def register_dataset(name):
    """Registers a ClassyDataset subclass.

    This decorator allows Classy Vision to instantiate a subclass of
    ClassyDataset from a configuration file, even if the class itself is not
    part of the Classy Vision framework. To use it, apply this decorator to a
    ClassyDataset subclass like this:

    |    @register_dataset("my_dataset")
    |    class MyDataset(ClassyDataset):
    |        ...

    To instantiate a dataset from a configuration file, see
    :func:`build_dataset`."""

    def register_dataset_cls(cls):
        if name in DATASET_REGISTRY:
            raise ValueError("Cannot register duplicate dataset ({})".format(name))
        if not issubclass(cls, ClassyDataset):
            raise ValueError(
                "Dataset ({}: {}) must extend ClassyDataset".format(name, cls.__name__)
            )
        if cls.__name__ in DATASET_CLASS_NAMES:
            raise ValueError(
                "Cannot register dataset with duplicate class name({})".format(
                    cls.__name__
                )
            )
        DATASET_REGISTRY[name] = cls
        DATASET_CLASS_NAMES.add(cls.__name__)
        return cls

    return register_dataset_cls


# automatically import any Python files in the dataset/ directory
import_all_modules(FILE_ROOT, "classy_vision.dataset")

from .classy_dataset import ClassyDataset  # isort:skip
from .classy_hmdb51 import HMDB51Dataset  # isort:skip
from .classy_kinetics400 import Kinetics400Dataset  # isort:skip
from .classy_synthetic_image import SyntheticImageDataset  # isort:skip
from .classy_synthetic_video import SyntheticVideoDataset  # isort:skip
from .classy_ucf101 import UCF101Dataset  # isort:skip
from .classy_video_dataset import ClassyVideoDataset  # isort:skip

__all__ = [
    "ClassyDataset",
    "ClassyVideoDataset",
    "HMDB51Dataset",
    "Kinetics400Dataset",
    "SyntheticImageDataset",
    "SyntheticVideoDataset",
    "UCF101Dataset",
    "build_dataset",
    "register_dataset",
]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

import torch
from classy_vision.generic.distributed_util import all_reduce_sum
from classy_vision.generic.util import is_pos_int
from classy_vision.meters import ClassyMeter

from . import register_meter


@register_meter("accuracy")
class AccuracyMeter(ClassyMeter):
    """Meter to calculate top-k accuracy for single label
       image classification task.
    """

    def __init__(self, topk):
        """
        args:
            topk: list of int `k` values.
        """
        assert isinstance(topk, list), "topk must be a list"
        assert len(topk) > 0, "topk list should have at least one element"
        assert [is_pos_int(x) for x in topk], "each value in topk must be >= 1"

        self._topk = topk

        # _total_* variables store running, in-sync totals for the
        # metrics. These should not be communicated / summed.
        self._total_correct_predictions_k = None
        self._total_sample_count = None

        # _curr_* variables store counts since the last sync. Only
        # these should be summed across workers and they are reset
        # after each communication
        self._curr_correct_predictions_k = None
        self._curr_sample_count = None

        # Initialize all values properly
        self.reset()

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "AccuracyMeter":
        """Instantiates a AccuracyMeter from a configuration.

        Args:
            config: A configuration for a AccuracyMeter.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A AccuracyMeter instance.
        """
        return cls(topk=config["topk"])

    @property
    def name(self):
        return "accuracy"

    def sync_state(self):
        # Communications
        self._curr_correct_predictions_k = all_reduce_sum(
            self._curr_correct_predictions_k
        )
        self._curr_sample_count = all_reduce_sum(self._curr_sample_count)

        # Store results
        self._total_correct_predictions_k += self._curr_correct_predictions_k
        self._total_sample_count += self._curr_sample_count

        # Reset values until next sync
        self._curr_correct_predictions_k.zero_()
        self._curr_sample_count.zero_()

    @property
    def value(self):
        # Return value based on the local state of meter which
        # includes the local sample count since last sync and the total global sample
        # count obtained at the last sync
        correct_predictions = {
            k: self._curr_correct_predictions_k[i]
            + self._total_correct_predictions_k[i]
            for i, k in enumerate(self._topk)
        }
        sample_count = self._total_sample_count + self._curr_sample_count
        return {
            "top_{}".format(k): (correct_predictions[k] / sample_count).item()
            if sample_count
            else 0.0
            for k in self._topk
        }

    def get_classy_state(self):
        """Contains the states of the meter.
        """
        return {
            "name": self.name,
            "top_k": self._topk,
            "total_correct_predictions": self._total_correct_predictions_k.clone(),
            "total_sample_count": self._total_sample_count.clone(),
            "curr_sample_count": self._curr_sample_count.clone(),
            "curr_correct_predictions_k": self._curr_correct_predictions_k.clone(),
        }

    def set_classy_state(self, state):
        assert (
            self.name == state["name"]
        ), "State name {state_name} does not match meter name {obj_name}".format(
            state_name=state["name"], obj_name=self.name
        )
        assert (
            self._topk == state["top_k"]
        ), "top-k of state {state_k} does not match object's top-k {obj_k}".format(
            state_k=state["top_k"], obj_k=self._topk
        )

        # Restore the state -- correct_predictions and sample_count.
        self.reset()
        self._total_correct_predictions_k = state["total_correct_predictions"].clone()
        self._total_sample_count = state["total_sample_count"].clone()
        self._curr_correct_predictions_k = state["curr_correct_predictions_k"].clone()
        self._curr_sample_count = state["curr_sample_count"].clone()

    def __repr__(self):
        return repr({"name": self.name, "value": self.value})

    def update(self, model_output, target, **kwargs):
        """
        args:
            model_output: tensor of shape (B, C) where each value is
                          either logit or class probability.
            target:       tensor of shape (B).
            Note: For binary classification, C=2.
        """
        # Due to dummy samples, in some corner cases, the whole batch could
        # be dummy samples, in that case we want to not update meters on that
        # process
        if model_output.shape[0] == 0:
            return
        _, pred = model_output.topk(max(self._topk), dim=1, largest=True, sorted=True)

        correct_predictions = pred.eq(target.unsqueeze(1).expand_as(pred))
        for i, k in enumerate(self._topk):
            self._curr_correct_predictions_k[i] += (
                correct_predictions[:, :k].float().sum().item()
            )
        self._curr_sample_count += model_output.shape[0]

    def reset(self):
        self._total_correct_predictions_k = torch.zeros(len(self._topk))
        self._total_sample_count = torch.zeros(1)
        self._curr_correct_predictions_k = torch.zeros(len(self._topk))
        self._curr_sample_count = torch.zeros(1)

    def validate(self, model_output_shape, target_shape):
        assert (
            len(model_output_shape) == 2
        ), "model_output_shape must be (B, C) \
            Found shape {}".format(
            model_output_shape
        )
        assert (
            len(target_shape) == 1
        ), "target_shape must be (B) \
            Found shape {}".format(
            target_shape
        )
        assert (
            max(self._topk) < model_output_shape[1]
        ), "k in top_k, for \
            accuracy_meter cannot be larger than num_classes: \
            {}".format(
            model_output_shape[1]
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import os

import torch
import torchvision
from classy_vision.generic.util import is_pos_int


def add_generic_args(parser):
    """
    Adds generic command-line arguments for convnet training / testing to parser.
    """
    parser.add_argument(
        "--config_file", type=str, help="path to config file for model", required=True
    )
    parser.add_argument(
        "--device",
        default=None,
        type=str,
        help="device to use: either 'cpu' or 'gpu'. If unspecified, will use GPU when available and CPU otherwise.",
    )
    parser.add_argument(
        "--num_workers",
        default=4,
        type=int,
        help="number of dataloading workers (default = 4)",
    )
    parser.add_argument(
        "--checkpoint_folder",
        default="",
        type=str,
        help="""folder to use for checkpoints:
                        epochal checkpoints are stored as model_<epoch>.torch,
                        latest epoch checkpoint is at checkpoint.torch""",
    )
    parser.add_argument(
        "--pretrained_checkpoint_folder",
        default="",
        type=str,
        help="""folder to use for pre-trained checkpoints:
                        epochal checkpoints are stored as model_<epoch>.torch,
                        latest epoch checkpoint is at checkpoint.torch,
                        checkpoint is used for fine-tuning task, and it will
                        not resume training from the checkpoint""",
    )
    parser.add_argument(
        "--checkpoint_period",
        default=1,
        type=int,
        help="""Checkpoint every x phases (default 1)""",
    )
    parser.add_argument(
        "--show_progress",
        default=False,
        action="store_true",
        help="shows progress bar during training / testing",
    )
    parser.add_argument(
        "--skip_tensorboard",
        default=False,
        action="store_true",
        help="do not perform tensorboard visualization",
    )
    parser.add_argument(
        "--visdom_server",
        default="",
        type=str,
        help="visdom server to use (default None)",
    )
    parser.add_argument(
        "--visdom_port",
        default=8097,
        type=int,
        help="port of visdom server (default = 8097)",
    )
    parser.add_argument(
        "--profiler",
        default=False,
        action="store_true",
        help="specify this argument to profile training code",
    )
    parser.add_argument(
        "--debug",
        default=False,
        action="store_true",
        help="specify this argument for debugging mode",
    )
    parser.add_argument(
        "--ignore_checkpoint_config",
        default=False,
        action="store_true",
        help="""specify this argument to ignore
                        the compatibility of the config (or lack of config) attached
                        to the checkpoint; this will allow mismatches between
                        the training specified in the config and the
                        actual training of the model""",
    )
    parser.add_argument(
        "--log_freq",
        default=5,
        type=int,
        help="Logging frequency for LossLrMeterLoggingHook (default 5)",
    )
    parser.add_argument(
        "--video_backend",
        default="pyav",
        type=str,
        help="torchvision video decoder backend (pyav or video_reader). Default pyav",
    )
    parser.add_argument(
        "--distributed_backend",
        default="none",
        type=str,
        help="""Distributed backend: either 'none' (for non-distributed runs)
             or 'ddp' (for distributed runs). Default none.""",
    )

    return parser


def check_generic_args(args):
    """
    Perform assertions on generic command-line arguments.
    """

    # check types and values:
    assert is_pos_int(args.num_workers), "incorrect number of workers"
    assert is_pos_int(args.visdom_port), "incorrect visdom port"
    assert (
        args.device is None or args.device == "cpu" or args.device == "gpu"
    ), "unknown device"

    # check that CUDA is available:
    if args.device == "gpu":
        assert torch.cuda.is_available(), "CUDA required to train on GPUs"

    # create checkpoint folder if it does not exist:
    if args.checkpoint_folder != "" and not os.path.exists(args.checkpoint_folder):
        os.makedirs(args.checkpoint_folder, exist_ok=True)
        assert os.path.exists(args.checkpoint_folder), (
            "could not create folder %s" % args.checkpoint_folder
        )

    # when in debugging mode, enter debugger upon error:
    if args.debug:
        import sys
        from classy_vision.generic.debug import debug_info

        sys.excepthook = debug_info

    # check visdom server name:
    if args.visdom_server != "":
        if args.visdom_server.startswith("https://"):
            print("WARNING: Visdom does not work over HTTPS.")
            args.visdom_server = args.visdom_server[8:]
        if not args.visdom_server.startswith("http://"):
            args.visdom_server = "http://%s" % args.visdom_server

    # set the right backend for torchvision:
    torchvision.set_image_backend("accimage")

    # return input arguments:
    return args


def get_parser():
    """
    Return a standard command-line parser.
    """
    parser = argparse.ArgumentParser(
        description="""Start a Classy Vision training job.

    This can be used for training on your local machine, using CPU or GPU, and
    for distributed training. This script also supports Tensorboard, Visdom and
    checkpointing."""
    )

    parser = add_generic_args(parser)
    return parser


def parse_train_arguments(parser=None):
    """
    Assert and parse the command-line arguments of a given (or default) parser.
    """

    # set input arguments:
    if parser is None:
        parser = get_parser()

    # parse input arguments:
    args = parser.parse_args()

    # assertions:
    args = check_generic_args(args)
    return args



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
import unittest.mock as mock
from test.generic.config_utils import get_test_classy_task, get_test_classy_video_task

from classy_vision.hooks import ProfilerHook


class TestProfilerHook(unittest.TestCase):
    @mock.patch("torch.autograd.profiler.profile", auto_spec=True)
    @mock.patch("classy_vision.hooks.profiler_hook.summarize_profiler_info")
    def test_profiler(
        self,
        mock_summarize_profiler_info: mock.MagicMock,
        mock_profile_cls: mock.MagicMock,
    ) -> None:
        """
        Tests that a profile instance is returned by the profiler
        and that the profiler actually ran.
        """
        mock_summarize_profiler_info.return_value = ""

        mock_profile = mock.MagicMock()
        mock_profile_returned = mock.MagicMock()
        mock_profile.__enter__.return_value = mock_profile_returned
        mock_profile_cls.return_value = mock_profile

        for task in [get_test_classy_task(), get_test_classy_video_task()]:
            task.prepare()
            local_variables = {}

            # create a model tensorboard hook
            profiler_hook = ProfilerHook()

            with self.assertLogs():
                profiler_hook.on_start(task, local_variables)

            # a new profile should be created with use_cuda=True
            mock_profile_cls.assert_called_once_with(use_cuda=True)
            mock_profile_cls.reset_mock()

            # summarize_profiler_info should have been called once with the profile
            mock_summarize_profiler_info.assert_called_once()
            profile = mock_summarize_profiler_info.call_args[0][0]
            mock_summarize_profiler_info.reset_mock()
            self.assertEqual(profile, mock_profile_returned)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import shutil
import tempfile
import unittest

import torch
from classy_vision.dataset import ClassyDataset, build_dataset
from classy_vision.dataset.image_path_dataset import ImagePathDataset
from torchvision import transforms


class TestImageDataset(unittest.TestCase):
    def get_test_image_dataset(self):
        config = {
            "name": "synthetic_image",
            "crop_size": 224,
            "num_channels": 3,
            "seed": 0,
            "class_ratio": 0.5,
            "num_samples": 100,
            "batchsize_per_replica": 1,
            "use_shuffle": False,
            "transforms": [
                {
                    "name": "apply_transform_to_key",
                    "transforms": [{"name": "ToTensor"}],
                    "key": "input",
                }
            ],
        }
        dataset = build_dataset(config)
        return dataset

    def setUp(self):
        # create a base directory to write image files to
        self.base_dir = tempfile.mkdtemp()
        os.mkdir(f"{self.base_dir}/0")
        os.mkdir(f"{self.base_dir}/1")

    def tearDown(self):
        # delete all the temporary data created
        shutil.rmtree(self.base_dir)

    def get_dataset_config(self):
        return {
            "batchsize_per_replica": 1,
            "use_shuffle": False,
            "num_samples": None,
            "transforms": [
                {
                    "name": "apply_transform_to_key",
                    "transforms": [{"name": "ToTensor"}],
                    "key": "input",
                }
            ],
        }

    @unittest.skip(
        "Skipping test since build_dataset doesn't "
        "work right now for ImagePathDataset"
    )
    def test_build_dataset(self):
        config = self.get_dataset_config()
        dataset = build_dataset(config)
        self.assertIsInstance(dataset, ClassyDataset)

    def test_image_dataset(self):
        image_paths = []
        inputs = []
        targets = []
        dataloader = self.get_test_image_dataset().iterator()
        for i, sample in enumerate(dataloader):
            input = sample["input"]
            target = sample["target"]
            image = transforms.ToPILImage()(input.squeeze())
            path = f"{self.base_dir}/{target.item()}/{i}.png"
            # save the image in a lossless format (png)
            image.save(path)
            image_paths.append(path)
            inputs.append(input)
            targets.append(target)

        # config for the image dataset
        config = self.get_dataset_config()

        # create an image dataset from the list of images
        dataset = ImagePathDataset.from_config(
            config, image_paths=image_paths, targets=targets
        )
        dataloader = dataset.iterator()
        # the samples should be in the same order
        for sample, expected_input, expected_target in zip(dataloader, inputs, targets):
            self.assertTrue(torch.allclose(sample["input"], expected_input))
            self.assertEqual(sample["target"], expected_target)

        # test the dataset works without targets as well
        dataset = ImagePathDataset.from_config(config, image_paths=image_paths)
        dataloader = dataset.iterator()
        # the samples should be in the same order
        for sample, expected_input in zip(dataloader, inputs):
            self.assertTrue(torch.allclose(sample["input"], expected_input))

        # create an image dataset from the root dir
        dataset = ImagePathDataset.from_config(config, image_paths=self.base_dir)
        dataloader = dataset.iterator()
        # test that we get the same class distribution
        # we don't test the actual samples since the ordering isn't defined
        counts = [0, 0]
        for sample in dataloader:
            counts[sample["target"].item()] += 1
        expected_counts = [0, 0]
        for target in targets:
            expected_counts[target.item()] += 1
        self.assertEqual(counts, expected_counts)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
import torch.nn as nn


class ClassyBlock(nn.Module):
    """
    This is a thin wrapper for head execution, which records the output of
    wrapped module for executing the heads forked from this module.
    """

    def __init__(self, name, module):
        super().__init__()
        self.name = name
        self.output = torch.zeros(0)
        self._module = module
        self._should_cache_output = False

    def set_cache_output(self, should_cache_output: bool = True):
        """
        Whether to cache the output of wrapped module for head execution.
        """
        self._should_cache_output = should_cache_output

    def forward(self, input):
        output = self._module(input)
        if self._should_cache_output:
            self.output = output
        return output



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

import torch
from classy_vision.models import ClassyModel, build_model


class TestMLPModel(unittest.TestCase):
    def test_build_model(self):
        config = {"name": "mlp", "input_dim": 3, "output_dim": 1, "hidden_dims": [2]}
        model = build_model(config)
        self.assertTrue(isinstance(model, ClassyModel))
        self.assertEqual(model.model_depth, 2)

        tensor = torch.tensor([[1, 2, 3]], dtype=torch.float)
        output = model.forward(tensor)
        self.assertEqual(output.shape, torch.Size([1, 1]))

        tensor = torch.tensor([[1, 2, 3], [1, 2, 3]], dtype=torch.float)
        output = model.forward(tensor)
        self.assertEqual(output.shape, torch.Size([2, 1]))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

import torch
from classy_vision.generic.util import is_pos_int
from classy_vision.meters import ClassyMeter
from classy_vision.meters.accuracy_meter import AccuracyMeter

from . import register_meter


@register_meter("video_accuracy")
class VideoAccuracyMeter(ClassyMeter):
    """Meter to calculate top-k video-level accuracy for single label
       video classification task. Video-level accuarcy is computed by averaging
       clip-level predictions and compare the reslt with video-level groundtruth
       label.
    """

    def __init__(self, topk, clips_per_video_train, clips_per_video_test):
        """
        args:
            topk: list of int `k` values.
            clips_per_video_train: No. of clips sampled per video at train time
            clips_per_video_test: No. of clips sampled per video at test time
        """
        assert isinstance(topk, list), "topk must be a list"
        assert len(topk) > 0, "topk list should have at least one element"
        assert [is_pos_int(x) for x in topk], "each value in topk must be >= 1"

        self._clips_per_video_train = clips_per_video_train
        self._clips_per_video_test = clips_per_video_test
        self._accuracy_meter = AccuracyMeter(topk)

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "VideoAccuracyMeter":
        """Instantiates a VideoAccuracyMeter from a configuration.

        Args:
            config: A configuration for a VideoAccuracyMeter.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A VideoAccuracyMeter instance.
        """
        return cls(
            topk=config["topk"],
            clips_per_video_train=config.get("clips_per_video_train", 1),
            clips_per_video_test=config["clips_per_video_test"],
        )

    @property
    def name(self):
        return "video_accuracy"

    @property
    def value(self):
        return self._accuracy_meter.value

    def sync_state(self):
        self._accuracy_meter.sync_state()

    def get_classy_state(self):
        """Contains the states of the meter.
        """
        state = {}
        state["accuracy_state"] = self._accuracy_meter.get_classy_state()
        state["name"] = "video_accuracy"
        state["clips_per_video_train"] = self._clips_per_video_train
        state["clips_per_video_test"] = self._clips_per_video_test
        return state

    def set_classy_state(self, state):
        assert (
            "video_accuracy" == state["name"]
        ), "State name {state_name} does not match meter name {obj_name}".format(
            state_name=state["name"], obj_name=self.name
        )
        assert (
            self._clips_per_video_train == state["clips_per_video_train"]
        ), "incompatible clips_per_video_train for video accuracy"
        assert (
            self._clips_per_video_test == state["clips_per_video_test"]
        ), "incompatible clips_per_video_test for video accuracy"
        # Restore the state -- correct_predictions and sample_count.
        self.reset()
        self._accuracy_meter.set_classy_state(state["accuracy_state"])

    def __repr__(self):
        return repr({"name": self.name, "value": self._accuracy_meter.value})

    def update(self, model_output, target, is_train, **kwargs):
        """
        args:
            model_output: tensor of shape (B * clips_per_video, C) where each value is
                          either logit or class probability.
            target:       tensor of shape (B * clips_per_video).
            is_train     if True, it is training stage when meter is updated

            Note: For binary classification, C=2.
        """
        num_clips = len(model_output)
        if num_clips == 0:
            # It is possible that a minibatch entirely contains dummy samples
            # when dataset is sharded. In such case, the effective target and output
            # can be empty, and we immediately return
            return

        clips_per_video = (
            self._clips_per_video_train if is_train else self._clips_per_video_test
        )
        assert num_clips % clips_per_video == 0, (
            "For video model testing, batch size must be a multplier of No. of "
            "clips per video"
        )
        num_videos = num_clips // clips_per_video
        for i in range(num_videos):
            clip_labels = target[i * clips_per_video : (i + 1) * clips_per_video]
            assert (
                len(torch.unique(clip_labels)) == 1
            ), "all clips from the same video should have same label"

        video_target = target[::clips_per_video]
        video_model_output = torch.mean(
            torch.reshape(model_output, (num_videos, clips_per_video, -1)), 1
        )
        self._accuracy_meter.update(video_model_output, video_target)

    def reset(self):
        self._accuracy_meter.reset()

    def validate(self, model_output_shape, target_shape):
        self._accuracy_meter.validate(model_output_shape, target_shape)



#!/usr/bin/env python3 -u
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import logging
import os
import socket
import subprocess
import sys
import time
from argparse import REMAINDER, ArgumentParser
from contextlib import closing

import ray


class NodeLaunchActor:
    """Ray actor. The code here will run in each node allocated by Ray."""

    def run(self, master_addr, master_port, node_rank, dist_world_size, args):
        processes = []

        # set PyTorch distributed related environmental variables
        current_env = os.environ.copy()
        current_env["MASTER_ADDR"] = master_addr
        current_env["MASTER_PORT"] = str(master_port)
        current_env["WORLD_SIZE"] = str(dist_world_size)

        if "OMP_NUM_THREADS" not in os.environ and args.nproc_per_node > 1:
            current_env["OMP_NUM_THREADS"] = str(1)
            print(
                "*****************************************\n"
                "Setting OMP_NUM_THREADS environment variable for each process "
                "to be {} in default, to avoid your system being overloaded, "
                "please further tune the variable for optimal performance in "
                "your application as needed. \n"
                "*****************************************".format(
                    current_env["OMP_NUM_THREADS"]
                )
            )

        # Set the init_method and rank of the process for distributed training.
        for local_rank in range(0, args.nproc_per_node):
            # each process's rank
            dist_rank = args.nproc_per_node * node_rank + local_rank
            current_env["RANK"] = str(dist_rank)
            current_env["LOCAL_RANK"] = str(local_rank)

            # spawn the processes
            with_python = not args.no_python
            cmd = []
            if with_python:
                cmd = [sys.executable, "-u"]
                if args.module:
                    cmd.append("-m")
            else:
                if not args.use_env:
                    raise ValueError(
                        "When using the '--no_python' flag, "
                        "you must also set the '--use_env' flag."
                    )
                if args.module:
                    raise ValueError(
                        "Don't use both the '--no_python' flag"
                        "and the '--module' flag at the same time."
                    )

            cmd.append(args.training_script)

            if not args.use_env:
                cmd.append("--local_rank={}".format(local_rank))

            cmd.extend(args.training_script_args)
            process = subprocess.Popen(cmd, env=current_env)
            processes.append(process)

        for process in processes:
            process.wait()
            if process.returncode != 0:
                raise subprocess.CalledProcessError(
                    returncode=process.returncode, cmd=cmd
                )

    def get_node_ip(self):
        return ray.services.get_node_ip_address()

    def find_free_port(self):
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(("", 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            return s.getsockname()[1]


def wait_for_gpus(world_size, timeout_secs=360):
    n_gpus = int(ray.cluster_resources().get("GPU", 0))
    elapsed_time = 0
    while n_gpus < world_size:
        logging.warning(
            f"Not enough GPUs available ({n_gpus} available,"
            f"need {world_size}), waiting 10 seconds"
        )
        time.sleep(10)
        elapsed_time += 10
        if elapsed_time > timeout_secs:
            raise RuntimeError("Timeout: could not find enough GPUs")
        n_gpus = int(ray.cluster_resources().get("GPU", 0))


def parse_args():
    """Helper function parsing the command line options.
    @retval ArgumentParser
    """
    parser = ArgumentParser(
        description="Classy Vision distributed training launch "
        "helper utility that will spawn up multiple nodes using Ray"
    )

    # Optional arguments for the launch helper
    parser.add_argument(
        "--nnodes",
        type=int,
        default=1,
        help="The number of nodes to use for distributed training",
    )
    parser.add_argument(
        "--nproc_per_node",
        type=int,
        default=1,
        help="The number of processes to launch on each node, "
        "for GPU training, this is recommended to be set "
        "to the number of GPUs in your system so that "
        "each process can be bound to a single GPU.",
    )
    parser.add_argument(
        "--use_env",
        default=False,
        action="store_true",
        help="Use environment variable to pass "
        "'local rank'."
        "If set to True, the script will not pass "
        "--local_rank as argument, and will instead set LOCAL_RANK.",
    )
    parser.add_argument(
        "-m",
        "--module",
        default=False,
        action="store_true",
        help="Changes each process to interpret the launch script "
        "as a python module, executing with the same behavior as"
        "'python -m'.",
    )
    parser.add_argument(
        "--no_python",
        default=False,
        action="store_true",
        help='Do not prepend the training script with "python" - just exec '
        "it directly. Useful when the script is not a Python script.",
    )
    # Ray-related arguments
    group = parser.add_argument_group("Ray related arguments")
    group.add_argument("--ray-address", default="auto", type=str)

    # positional
    parser.add_argument(
        "training_script",
        type=str,
        help="The full path to the single GPU training "
        "program/script to be launched in parallel, "
        "followed by all the arguments for the "
        "training script",
    )

    # rest from the training program
    parser.add_argument("training_script_args", nargs=REMAINDER)
    return parser.parse_args()


def main():
    args = parse_args()

    ray.init(address=args.ray_address)

    # world size in terms of number of processes
    dist_world_size = args.nproc_per_node * args.nnodes

    wait_for_gpus(dist_world_size)

    # Set up Ray distributed actors.
    actor = ray.remote(num_cpus=1, num_gpus=args.nproc_per_node)(NodeLaunchActor)
    workers = [actor.remote() for i in range(args.nnodes)]

    # Set worker 0 as the master
    master_addr = ray.get(workers[0].get_node_ip.remote())
    master_port = ray.get(workers[0].find_free_port.remote())

    unfinished = [
        worker.run.remote(
            master_addr=master_addr,
            master_port=master_port,
            node_rank=i,
            dist_world_size=dist_world_size,
            args=args,
        )
        for i, worker in enumerate(workers)
    ]

    try:
        while len(unfinished) > 0:
            finished, unfinished = ray.wait(unfinished)
            finished = ray.get(finished)
    except Exception as inst:
        logging.exception("An error occurred:")

    ray.shutdown()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

from . import ClassyParamScheduler, UpdateInterval, register_param_scheduler
from .step_scheduler import StepParamScheduler


@register_param_scheduler("step_with_fixed_gamma")
class StepWithFixedGammaParamScheduler(ClassyParamScheduler):
    """
    Decays the param value by gamma at equal number of steps so as to have the
    specified total number of decays.

    Example:
      base_lr: 0.1
      gamma: 0.1
      num_decays: 3
      num_epochs: 120

    Then the param value will be 0.1 for epochs 0-29, 0.01 for
    epochs 30-59, 0.001 for epoch 60-89, 0.0001 for epochs 90-119.
    """

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "StepWithFixedGammaParamScheduler":
        """Instantiates a StepWithFixedGammaParamScheduler from a configuration.

        Args:
            config: A configuration for a StepWithFixedGammaParamScheduler.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A StepWithFixedGammaParamScheduler instance.
        """
        for key in ["base_lr", "gamma", "num_decays", "num_epochs"]:
            assert key in config, f"Step with fixed decay scheduler requires: {key}"
        for key in ["base_lr", "gamma"]:
            assert (
                isinstance(config[key], (int, float)) and config[key] > 0
            ), f"{key} must be a positive number"
        for key in ["num_decays", "num_epochs"]:
            assert (
                isinstance(config[key], int) and config[key] > 0
            ), f"{key} must be a positive integer"

        return cls(
            base_lr=config["base_lr"],
            num_decays=config["num_decays"],
            gamma=config["gamma"],
            num_epochs=config["num_epochs"],
        )

    def __init__(self, base_lr, num_decays, gamma, num_epochs):
        super().__init__()

        self.base_lr = base_lr
        self.num_decays = num_decays
        self.gamma = gamma
        self.num_epochs = num_epochs
        values = [base_lr]
        for _ in range(num_decays):
            values.append(values[-1] * gamma)

        self._step_param_scheduler = StepParamScheduler(
            num_epochs=num_epochs, values=values
        )

        # make this a STEP scheduler
        self.update_interval = UpdateInterval.STEP

    def __call__(self, where: float) -> float:
        return self._step_param_scheduler(where)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Optional

import torch
from classy_vision.generic.distributed_util import barrier, is_distributed_training_run
from classy_vision.hooks import ClassyHookFunctions
from classy_vision.tasks import ClassyTask


class ClassyTrainer:
    """Base class for shared training code.

    A trainer is responsible for setting up the environment for
    training, for instance: configuring rendezvous for distributed
    training, deciding what GPU to use and so on. Trainers also
    control the outer portion of the training loop, but delegate to
    the task to decide how exactly to perform inference, compute loss
    etc. That allows combining tasks with different trainers depending
    on whether you want to train on your current machine, AWS cluster
    etc.

    """

    def __init__(
        self,
        use_gpu: Optional[bool] = None,
        num_dataloader_workers: int = 0,
        dataloader_mp_context: Optional[str] = None,
    ):
        """Constructor for ClassyTrainer.

        Args:
            use_gpu: If true, then use GPUs for training.
                If None, then check if we have GPUs available, if we do
                then use GPU for training.
            num_dataloader_workers: Number of CPU processes doing dataloading
                per GPU. If 0, then dataloading is done on main thread.
            dataloader_mp_context: Determines how to launch
                new processes for dataloading. Must be one of "fork", "forkserver",
                "spawn". If None, process launching is inherited from parent.
        """
        if use_gpu is None:
            use_gpu = torch.cuda.is_available()
        self.use_gpu = use_gpu
        self.num_dataloader_workers = num_dataloader_workers
        self.dataloader_mp_context = dataloader_mp_context

    def train(self, task: ClassyTask):
        """Runs training phases, phases are generated from the config.

        Args:
            task: Task to be used in training. It should contain
                everything that is needed for training
        """

        pin_memory = self.use_gpu and torch.cuda.device_count() > 1
        task.prepare(
            num_dataloader_workers=self.num_dataloader_workers,
            pin_memory=pin_memory,
            use_gpu=self.use_gpu,
            dataloader_mp_context=self.dataloader_mp_context,
        )
        assert isinstance(task, ClassyTask)

        if is_distributed_training_run():
            task.init_distributed_data_parallel_model()

        local_variables = {}
        task.run_hooks(local_variables, ClassyHookFunctions.on_start.name)

        while not task.done_training():
            task.advance_phase()

            # Start phase hooks
            task.run_hooks(local_variables, ClassyHookFunctions.on_phase_start.name)
            while True:
                # Process next sample
                try:
                    task.train_step(self.use_gpu, local_variables)
                except StopIteration:
                    break

            logging.info("Syncing meters on phase end...")
            for meter in task.meters:
                meter.sync_state()
            logging.info("...meters synced")
            barrier()
            task.run_hooks(local_variables, ClassyHookFunctions.on_phase_end.name)

        task.run_hooks(local_variables, ClassyHookFunctions.on_end.name)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import re
import unittest
import unittest.mock as mock
from itertools import product
from test.generic.config_utils import get_test_classy_task

from classy_vision.generic.perf_stats import PerfStats
from classy_vision.hooks import TimeMetricsHook


class TestTimeMetricsHook(unittest.TestCase):
    @mock.patch("time.time")
    @mock.patch("classy_vision.hooks.time_metrics_hook.PerfStats.report_str")
    @mock.patch("classy_vision.hooks.time_metrics_hook.get_rank")
    def test_time_metrics(
        self,
        mock_get_rank: mock.MagicMock,
        mock_report_str: mock.MagicMock,
        mock_time: mock.MagicMock,
    ) -> None:
        """
        Tests that the progress bar is created, updated and destroyed correctly.
        """
        rank = 5
        mock_get_rank.return_value = rank

        mock_report_str.return_value = ""
        local_variables = {}

        for log_freq, train in product([5, None], [True, False]):
            # create a time metrics hook
            time_metrics_hook = TimeMetricsHook(log_freq=log_freq)

            phase_type = "train" if train else "test"

            task = get_test_classy_task()
            task.prepare()
            task.train = train

            # on_phase_start() should set the start time and perf_stats
            start_time = 1.2
            mock_time.return_value = start_time
            time_metrics_hook.on_phase_start(task, local_variables)
            self.assertEqual(time_metrics_hook.start_time, start_time)
            self.assertTrue(isinstance(local_variables.get("perf_stats"), PerfStats))

            # test that the code doesn't raise an exception if losses is empty
            try:
                time_metrics_hook.on_phase_end(task, local_variables)
            except Exception as e:
                self.fail("Received Exception when losses is []: {}".format(e))

            # check that _log_performance_metrics() is called after on_loss_and_meter()
            # every log_freq batches and after on_phase_end()
            with mock.patch.object(
                time_metrics_hook, "_log_performance_metrics"
            ) as mock_fn:
                num_batches = 20

                for i in range(num_batches):
                    task.losses = list(range(i))
                    time_metrics_hook.on_loss_and_meter(task, local_variables)
                    if log_freq is not None and i and i % log_freq == 0:
                        mock_fn.assert_called_with(task, local_variables)
                        mock_fn.reset_mock()
                        continue
                    mock_fn.assert_not_called()

                time_metrics_hook.on_phase_end(task, local_variables)
                mock_fn.assert_called_with(task, local_variables)

            task.losses = [0.23, 0.45, 0.34, 0.67]

            end_time = 10.4
            avg_batch_time_ms = 2.3 * 1000
            mock_time.return_value = end_time

            # test _log_performance_metrics()
            with self.assertLogs() as log_watcher:
                time_metrics_hook._log_performance_metrics(task, local_variables)

            # there should 2 be info logs for train and 1 for test
            self.assertEqual(len(log_watcher.output), 2 if train else 1)
            self.assertTrue(
                all(
                    log_record.levelno == logging.INFO
                    for log_record in log_watcher.records
                )
            )
            match = re.search(
                (
                    r"Average {} batch time \(ms\) for {} batches: "
                    r"(?P<avg_batch_time>[-+]?\d*\.\d+|\d+)"
                ).format(phase_type, len(task.losses)),
                log_watcher.output[0],
            )
            self.assertIsNotNone(match)
            self.assertAlmostEqual(
                avg_batch_time_ms, float(match.group("avg_batch_time")), places=4
            )
            if train:
                self.assertIn(
                    f"Train step time breakdown (rank {rank})", log_watcher.output[1]
                )

            # if on_phase_start() is not called, 2 warnings should be logged
            # create a new time metrics hook
            local_variables = {}
            time_metrics_hook_new = TimeMetricsHook()

            with self.assertLogs() as log_watcher:
                time_metrics_hook_new.on_phase_end(task, local_variables)

            self.assertEqual(len(log_watcher.output), 2)
            self.assertTrue(
                all(
                    log_record.levelno == logging.WARN
                    for log_record in log_watcher.records
                )
            )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, List, Optional

import torch
from torch import Tensor

from . import ClassyLoss, build_loss, register_loss


@register_loss("sum_arbitrary")
class SumArbitraryLoss(ClassyLoss):
    """
    Sums a collection of (weighted) torch.nn losses.

    NOTE: this applies all the losses to the same output and does not support
    taking a list of outputs as input.
    """

    def __init__(self, losses: List[float], weights: Optional[Tensor] = None) -> None:
        super().__init__()
        if weights is None:
            weights = torch.ones((len(losses)))

        self.losses = losses
        self.weights = weights

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "SumArbitraryLoss":
        """Instantiates a SumArbitraryLoss from a configuration.

        Args:
            config: A configuration for a SumArbitraryLoss.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A SumArbitraryLoss instance.
        """
        assert (
            type(config["losses"]) == list and len(config["losses"]) > 0
        ), "losses must be a list of registered losses with length > 0"
        assert type(config["weights"]) == list and len(config["weights"]) == len(
            config["losses"]
        ), "weights must be None or a list and have same length as losses"

        loss_modules = []
        for loss_config in config["losses"]:
            loss_modules.append(build_loss(loss_config))

        assert all(
            isinstance(loss_module, ClassyLoss) for loss_module in loss_modules
        ), "All losses must be registered, valid ClassyLosses"

        return cls(losses=loss_modules, weights=config.get("weights", None))

    def forward(self, prediction, target):
        for idx, loss in enumerate(self.losses):
            current_loss = loss(prediction, target)
            if idx == 0:
                total_loss = current_loss
            else:
                total_loss = total_loss.add(self.weights[idx], current_loss)
        return total_loss
